--[[
Name: WhoLib-1.0
Revision: $Revision: 34545 $
Author(s): ALeX Kazik (alx@kazik.de)
Website: http://wowace.com/wiki/WhoLib-1.0
Documentation: http://wowace.com/wiki/WhoLib-1.0
SVN: svn://dev.wowace.com/wowace/trunk/WhoLib/
Description: Queing of /who and SendWho() queries and a much better interface, with gurantee to be executed & callback.
License: http://creativecommons.org/licenses/by-nc-sa/2.5/
Dependencies: AceLibrary, AceEvent-2.0, AceHook-2.1, AceOO-2.0, AceLocale-2.2, Deformat
]]

local MAJOR_VERSION = 'WhoLib-1.0'
local MINOR_VERSION = '$Revision: 34545 $'

if not AceLibrary then error(MAJOR_VERSION .. ' requires AceLibrary') end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

if not AceLibrary:HasInstance('AceEvent-2.0') then error(MAJOR_VERSION .. ' requires AceEvent-2.0.') end
if not AceLibrary:HasInstance('AceHook-2.1') then error(MAJOR_VERSION .. ' requires AceHook-2.1') end
if not AceLibrary:HasInstance('AceOO-2.0') then error(MAJOR_VERSION .. ' requires AceOO-2.0') end
if not AceLibrary:HasInstance('AceLocale-2.2') then error(MAJOR_VERSION .. ' requires AceLocale-2.2') end

local AceOO = AceLibrary:GetInstance('AceOO-2.0')
local lib = AceOO.Mixin {
				'Who',
				'WHO_QUEUE_USER',
				'WHO_QUEUE_QUIET',
				'WHO_QUEUE_BACKGROUND',
				'SetDebug',
			}

-- externals

function lib.Who(self, query, queue, callback)
	-- self is the object which calls us
	
	if(queue == WhoImproved.WHO_QUEUE_USER)then
		if(WhoFrame:IsVisible())then
			WhoImproved:GuiWho(query)
		else
			WhoImproved:ConsoleWho(query)
		end
	else
		if(queue ~= WhoImproved.WHO_QUEUE_QUIET)then
			queue = WhoImproved.WHO_QUEUE_BACKGROUND
		end
		WhoImproved:AskWho(query, queue, self, callback)
	end
end

function lib.SetDebug(self, mode)
	-- self is the object which calls us
	
	WhoImproved.Debug = mode
end

-- internals

function lib:AskWhoNext()
	local queue, query = nil, nil
	for k,v in ipairs(self.Queue) do
		if(WhoFrame:IsVisible() and k > self.WHO_QUEUE_SENDWHO)then
			break
		end
		if(#v > 0)then
			queue = k
			query = tremove(v, 1)
			break
		end
	end
	if(query)then
		self.WhoInProgress = true
		self.Result = {}
		self.Query = query
		self.Total = -1
		if(query.c_show == true)then
			DEFAULT_CHAT_FRAME:AddMessage(string.format(self.L['console_query'], query.q))
		end
		if(queue == self.WHO_QUEUE_USER)then
			WhoFrameEditBox:SetText(query.q)
		end
		if(query.m == self.WHO_QUEUE_QUIET or query.m == self.WHO_QUEUE_BACKGROUND)then
			SetWhoToUI(1)
			self.Quiet = true
		elseif(query.gui == true)then
			SetWhoToUI(1)
		else
			SetWhoToUI(0)
		end
		self.hooks.SendWho(query.q)
	else
		self.WhoInProgress = false
	end
end

function lib:AskWho(query, queue, handler, callback, opts)
	if(type(opts) ~= 'table')then
		opts = {}
	end
	opts.q = query
	opts.h = handler
	opts.c = callback
	opts.m = queue
	tinsert(self.Queue[queue], opts)
	if(self.Debug)then
		DEFAULT_CHAT_FRAME:AddMessage('WhoLib: [' .. opts.m .. '] added "' .. opts.q .. '", queues=' .. #self.Queue[1] .. ', '.. #self.Queue[2] .. ', '.. #self.Queue[3] .. ', '.. #self.Queue[4])
	end
	if(not self.WhoInProgress)then
		self:AskWhoNext()
	end
end

function lib:ReturnWho()
	if(self.Query.m == self.WHO_QUEUE_QUIET or self.Query.m == self.WHO_QUEUE_BACKGROUND)then
		self.Quiet = nil
	end
	if(self.Debug)then
		DEFAULT_CHAT_FRAME:AddMessage('WhoLib: [' .. self.Query.m .. '] returned "' .. self.Query.q .. '", total=' .. self.Total ..' , queues=' .. #self.Queue[1] .. ', '.. #self.Queue[2] .. ', '.. #self.Queue[3] .. ', '.. #self.Queue[4])
	end
	if(type(self.Query.c) == 'function')then
		self.Query.c(self.Query.q, self.Total, self.Result)
	elseif(type(self.Query.c) == 'string')then
		self.Query.h[self.Query.c](self.Query.h, self.Query.q, self.Total, self.Result)
	end
	self:TriggerEvent('WHO_QUERY_RESULT', self.Query.q, self.Total, self.Result)
	self:ScheduleEvent(self.AskWhoNext, 5, self)
end

function lib:GuiWho(msg)
	if(msg == self.L['gui_wait'])then
		return
	end

	local q1count = #self.Queue[self.WHO_QUEUE_USER]
	for _,v in pairs(self.Queue[self.WHO_QUEUE_USER]) do
		if(v.gui == true)then
			return
		end
	end
	if(self.WhoInProgress)then
		WhoFrameEditBox:SetText(self.L['gui_wait'])
	end
	self:AskWho(msg, self.WHO_QUEUE_USER, nil, nil, {gui = true})
	WhoFrameEditBox:ClearFocus();
end

function lib:ConsoleWho(msg)
	WhoFrameEditBox:SetText(msg)
	local c_show = false
	local q1count = #self.Queue[self.WHO_QUEUE_USER]
	if(q1count > 0 and self.Queue[self.WHO_QUEUE_USER][q1count][q] == msg)then -- last query is itdenical: drop
		return
	end
	if(q1count == 1 and self.Queue[self.WHO_QUEUE_USER][1].c_show == false)then -- display 'queued' if console and not yet shown
		DEFAULT_CHAT_FRAME:AddMessage(string.format(self.L['console_queued'], msg))
		self.Queue[self.WHO_QUEUE_USER][1].c_show = true
	end
	if(q1count > 0)then
		DEFAULT_CHAT_FRAME:AddMessage(string.format(self.L['console_queued'], msg))
		c_show = true
	end
	self:AskWho(msg, self.WHO_QUEUE_USER, nil, nil, {c_show = c_show} )
end

-- hooks

function lib:SendWho(msg)
	self:AskWho(msg, self.WHO_QUEUE_SENDWHO)
end

function lib:WhoFrameEditBox_OnEnterPressed()
	self:GuiWho(WhoFrameEditBox:GetText())
end

-- new '/who' function

local function SlashWHO(msg)
	if(msg == '')then
		WhoImproved:GuiWho(WhoFrame_GetDefaultWhoCommand())
	elseif(WhoFrame:IsVisible())then
		WhoImproved:GuiWho(msg)
	else
		WhoImproved:ConsoleWho(msg)
	end
end

-- /wholibdebug: debug on/off; I decided not to use the AceConsole, since it's not neesed and requies less memory

local function SlashDebug()
	WhoImproved.Debug = not WhoImproved.Debug
	DEFAULT_CHAT_FRAME:AddMessage('WhoLib: Debugging is now ' .. (WhoImproved.Debug and 'on' or 'off'))
end

-- events

function lib:CHAT_MSG_SYSTEM(msg)
	local pla, _, lvl, rac, cla, gui, zon = self.Deformat:Deformat(arg1, WHO_LIST_GUILD_FORMAT)
	if(pla ~= nil)then
		tinsert(self.Result, {n=pla, g=gui, l=lvl, r=rac, c=cla, z=zon})
	else
		local pla, _, lvl, rac, cla, zon = self.Deformat:Deformat(arg1, WHO_LIST_FORMAT)
		if(pla ~= nil)then
			tinsert(self.Result, {n=pla, g='', l=lvl, r=rac, c=cla, z=zon})
		else
			local numWhoResults = self.Deformat:Deformat(arg1, WHO_NUM_RESULTS) or self.Deformat:Deformat(arg1, WHO_NUM_RESULTS_P1)
			if(numWhoResults ~= nil)then
				self.Total = numWhoResults
				self:ReturnWho()
			end
		end
	end
end

function lib:WHO_LIST_UPDATE()
	local num
	self.Total, num = GetNumWhoResults()
	for i=1, num do
		local charname, guildname, level, race, class, zone = GetWhoInfo(i)
		self.Result[i] = {n=charname, g=guildname, l=level, r=race, c=class, z=zone}
	end
	
	self:ReturnWho()
end

-- init

local function activate(self, oldLib, oldDeactivate)
	self.Queue = oldLib and oldLib.Queue or {[1]={}, [2]={}, [3]={}, [4]={}}
	self.SlashWho = oldLib and oldLib.SlashWho or SlashCmdList['WHO']
	self.WhoInProgress = oldLib and oldLib.WhoInProgress or false
	self.Result = oldLib and oldLib.Result or nil
	self.Query = oldLib and oldLib.Query or nil
	self.Total = oldLib and oldLib.Total or nil
	self.Quiet = oldLib and oldLib.Quiet or nil
	self.Debug = oldLib and oldLib.Debug or false
	
	SlashCmdList['WHO'] = SlashWHO
	
	SlashCmdList['WHOLIB_DEBUG'] = SlashDebug
	SLASH_WHOLIB_DEBUG1 = '/wholibdebug'
	
	if(self.activate)then
		self:activate(oldLib, oldDeactivate)
	end
	
	WhoImproved = self
	
	self.L = AceLibrary('AceLocale-2.2'):new('WhoLib')
	
	self.L:RegisterTranslations('enUS', function() return {
		['console_queued'] = 'Added "/who %s" to queue',
		['console_query'] = 'Result of "/who %s"',
		['gui_wait'] = '- Please Wait -',
	} end )
	
	self.WHO_QUEUE_USER = 1
	self.WHO_QUEUE_SENDWHO = 2
	self.WHO_QUEUE_QUIET = 3
	self.WHO_QUEUE_BACKGROUND = 4
	
	if oldDeactivate then
		oldDeactivate(oldLib)
	end
end

function lib:FriendsFrame_OnEvent()
	if(not (event == 'WHO_LIST_UPDATE' and self.Quiet))then
		self.hooks.FriendsFrame_OnEvent(event)
	end
end

function lib:CloseWhoFrame()
	if(not self.WhoInProgress)then
		self:AskWhoNext()
	end
end

local function external(self, major, instance)
	if(major == 'AceEvent-2.0')then
		instance:embed(self)
		self:UnregisterAllEvents()
		self:CancelAllScheduledEvents()
		self:RegisterEvent('CHAT_MSG_SYSTEM')
		self:RegisterEvent('WHO_LIST_UPDATE')
	elseif(major == 'AceHook-2.1')then
		instance:embed(self)
		self:UnhookAll()
		self:Hook('SendWho', true)
		self:Hook('WhoFrameEditBox_OnEnterPressed', true)
		self:Hook("FriendsFrame_OnEvent", true)
		self:SecureHook(WhoFrame, 'Hide', 'CloseWhoFrame')
	elseif(major == 'Deformat-2.0')then
		self.Deformat = instance
	end
end

AceLibrary:Register(lib, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
